/**
 * @file Widget组件
 * @author Yern Sun <sunyueran@baidu.com>
 */
(function () {
    // 默认自动监听事件列表
    var eventsMap = [
        'click',
        'touchend',
        'touchstart',
        'change'
    ];

    function Widget(instance) {
        var self = this;
        // 组件事件列表
        self.$handle = {};
        var widgetID = new Date().getTime() % 10000000 + Math.random();

        /**
         * 生成事件阻止信号，同时避免事件监听被多次调用
         * @return {number} 组件事件阻止token
         */
        self.$preventEvent = (function () {
            var token = +new Date;
            return function () {
                return token;
            };
        })();
        // 可直接选择dom
        if (self.$el && self.$el.length) {
            self.$el = $(self.$el);
        }
        else if (self.name) {
            // 通过widget注册dom
            // 避免同名组件dom被多次监听及事件污染
            self.$el = $('[widget="' + self.name + '"]').not('[widget-id]').eq(0);
        }

        // 依据新实例注册事件，并且避免由于继承导致的多次事件监听
        if (!canRegisterEvent()) {
            return;
        }

        function canRegisterEvent() {
            return instance === 'instance' && self.$el && self.$el.size() && !self.$el.attr('widget-id');
        }
        self.$el.attr('widget-id', widgetID);
        // 自动监听事件
        eventsMap.forEach(function (event) {
            self.$el.on(event, '[data-' + event + ']', function (el) {
                el.stopPropagation();
                var action = $(this).data(event);
                var result;
                var params = $(this).data('params') || $(this).data();

                if (!params instanceof Object) {
                    try {
                        params = JSON.parse(params);
                    }
                    catch (e) {
                        console.warn($(this)[0], '参数格式错误');
                        params = {};
                    }
                }

                // 触发代理事件
                if ($.isFunction(self[action])) {
                    result = self[action].call(self, el, params);
                }
                else {
                    console.warn('组件:(' + self.name + ')方法' + action + '不存在');
                }
                // 提供事件截取的接口
                if (result === self.$preventEvent()) {
                    return;
                }
                // 触发外部接口对应的内部事件
                (self.$handle[action] || []).forEach(function (handle) {
                    if ($.isFunction(handle)) {
                        handle.call(self, result, params);
                    }
                });
            });
        });


        // 实现自定义初始化函数，避免多次执行，将初始化函数放在实例创建时
        //        if (self.init && $.isFunction(self.init)) {
        //            self.init.apply(self, arguments);
        //        }
    }



    /**
     * ajax异步方法
     * @return {Object} promise对象
     */
    Widget.prototype.ajax = function () {
        var dfd = $.Deferred();
        return dfd;
    };

    /**
     * 加载pagelet
     * @param   {string} name       [[Description]]
     * @param   {Object}   [option={] 请求option
     * @return {Object} promise对象
     */
    Widget.prototype.load = function (name, option) {
        var dfd = $.Deferred();
        var callback = function () {};
        option = option || {};

        if (typeof option.cb === 'function') {
            callback = option.cb;
        }
        window.BNMJS.page.getData(function (data) {
            if (option.param instanceof Object || option.params instanceof Object) {
                option.param = toForm($.extend({}, data, option.param || option.params)) || '';
            }
            if (!option.hideLoading) {
                $('#loading').removeClass('hide');
                setTimeout(function () {
                    if (!$('#loading').hasClass('hide')) {
                        $('#loading').addClass('hide');
                        window.BNMJS.ui.toast.show('请求超时！');
                    }
                }, 15000);
            }

            window.BigPipe.load($.extend(
                option, {
                    pagelets: [name],
                    cb: function () {
                        $('#loading').addClass('hide');
                        dfd.resolve.apply(dfd.resolve, arguments);
                        callback.apply(this, arguments);

                    }
                }));
        });
        return dfd;
    };


    /**
     * 析构函数，注销组件事件，资源回收
     */
    Widget.prototype.finalize = function () {
        var self = this;
        if (self.$el) {
            self.$el.off();
            for (var k in self) {
                self[k] = null;
                delete self[k];
            }
        }
    };

    /**
     * 注册引用事件
     * @param {string} event  事件名称
     * @param {function} handle 事件handle
     */
    Widget.prototype.on = function (event, handle) {
        if (!this.$handle[event]) {
            this.$handle[event] = [];
        }
        this.$handle[event].push(handle);
        return this;
    };

    /**
     * 注销引用事件
     * @param {string} event 事件名称
     */
    Widget.prototype.off = function (event) {
        if (this.$handle[event]) {
            delete this.$handle[event];
        }
    };

    Widget.prototype.onMessage = function () {

    };

    /**
     * 组件继承方法
     * @return {Object} Widget实例
     */
    Widget.extend = function (obj) {
        var Super = this;
        var subClass;
        var $super = new Super('extend');

        function SubClass() {
            Super.apply(this, arguments);
            this.$superInstance = $super;
        }
        // 实现继承
        SubClass.prototype = new Super('extend');

        $.extend(SubClass.prototype, obj);

        SubClass.extend = Super.extend;

        SubClass.new = function (params) {
            subClass = new this('instance');
            // 多重继承待测试
            subClass.$super = function (method, args) {
                if (method in $super && $.isFunction($super[method])) {
                    var superInstance = $super;
                    while (superInstance && superInstance[method] === this[method]) {
                        superInstance = $super.$superInstance;
                    }
                    if (superInstance && superInstance[method]) {
                        superInstance[method].apply(this, args);
                    }
                }
            };
            // 避免init继承的过程中被多次调用，如需调用父类构造函数，请手动调用$super
            if (subClass.init && $.isFunction(subClass.init)) {
                subClass.init.apply(subClass, arguments);
            }

            return subClass;
        };

        // subClass = new SubClass();
        // subClass.super = Super.prototype;

        return SubClass;
    };


    Widget.new = function (widget) {
        return;
    };

    window.Widget = Widget;

    /**
     * json转queryString，暂时不支持深层次对象
     * @param   {Object} [data={}] json对象
     * @return {string} queryString
     */
    function toForm(data) {
        var rs = [];
        data = data || {};
        for (var k in data) {
            if (data[k] === null || data[k] === undefined) {
                data[k] = '';
            }
            rs.push(encodeURIComponent(k) + '=' + encodeURIComponent(data[k]));
        }
        return rs.join('&');
    }
})();